/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef prpriv_h___
#define prpriv_h___

/*
 * NSPR 2.0 Private API
 */

#include "private/pprio.h"
#include "private/pprthred.h"

#endif /* prpriv_h___ */
